"""
functionality to create the feature data necessary.
"""
from typing import List, Optional, Union, Tuple
from datetime import datetime, timedelta

import pandas as pd
import pytz

from ts_forecasting_pipeline.speccing import ModelSpecs
from ts_forecasting_pipeline.utils.time_utils import get_closest_quarter


def construct_features(
    time_range: Union[Tuple[datetime, datetime], datetime, str], specs: ModelSpecs
) -> pd.DataFrame:
    """
    Create a DataFrame based on ModelSpecs. This means loading data, potentially running a custom transformation,
    adding lags & returning only the relevant date rows.

    time_range can be specified by you (a range of steps between two datetimes or a single step - one datetime-,
    which is useful when we are predicting said step) but usually you’d say “train” or “test” to let the model specs
    determine the time steps.
    """
    datetime_indices = get_time_steps(time_range, specs)

    df = pd.DataFrame()

    # load raw data series for the outcome data
    df[specs.outcome_var.name] = specs.outcome_var.load_series()

    # perform the custom transformation, if needed, and store the transformation parameters to be able to back-transform
    if specs.transformation is not None:
        df[specs.outcome_var.name] = specs.transformation.transform(
            df[specs.outcome_var.name]
        )

    # load raw data series for the regressors
    for reg_spec in specs.regressors:
        df[reg_spec.name] = reg_spec.load_series()

    # perform the custom transformation, if needed
    if specs.regressor_transformation is not None:
        for reg_spec in specs.regressors:
            if reg_spec.name in specs.regressor_transformation:
                if specs.regressor_transformation[reg_spec.name] is not None:
                    df[reg_spec.name] = specs.regressor_transformation[
                        reg_spec.name
                    ].transform(df[reg_spec.name])

    # add lags on the outcome var
    df = add_lags(df, specs.outcome_var.name, specs.lags)
    outcome_lags = [
        lag_name
        for lag_name in [
            specs.outcome_var.name + "_" + lag_to_suffix(lag) for lag in specs.lags
        ]
        if lag_name in df.columns
    ]

    # now select only relevant columns and relevant datetime indices
    relevant_columns = (
        [specs.outcome_var.name] + outcome_lags + [r.name for r in specs.regressors]
    )
    df = df[relevant_columns].reindex(datetime_indices)

    # Check for nan values in lagged and regressor data
    if df.drop(specs.outcome_var.name, axis=1).isnull().values.any():
        raise Exception(
            "I found nan values and I obviously can't make forecasts lacking data. Here's how many I found:\n\n%s\n\n"
            % df.drop(specs.outcome_var.name, axis=1).isnull().sum()
        )

    # if specs.model_type == "OLS":
    #    df = sm.add_constant(df)

    return df


def get_time_steps(
    time_range: Union[str, datetime, Tuple[datetime, datetime]], specs: ModelSpecs
) -> pd.DatetimeIndex:
    """ get relevant datetime indices to build features for."""
    if not (
        isinstance(time_range, datetime)
        or (
            isinstance(time_range, tuple)
            and isinstance(time_range[0], datetime)
            and isinstance(time_range[1], datetime)
        )
        or (isinstance(time_range, str) and time_range in ("train", "test"))
    ):
        raise Exception(
            "Purpose for dataframe construction needs to be either 'train', 'test',"
            "a tuple of two datetime objects or one datetime object."
        )

    if isinstance(time_range, datetime):
        return pd.date_range(time_range, time_range, closed="left", freq="15T")
    elif isinstance(time_range, tuple):
        return pd.date_range(time_range[0], time_range[1], closed="left", freq="15T")

    length_of_data = specs.end_of_testing - specs.start_of_training
    datetime_indices = None
    if time_range == "train":
        end_of_training = get_closest_quarter(
            specs.start_of_training + length_of_data * specs.ratio_training_testing_data
        )
        # print("Start of training: %s" % specs.start_of_training)
        # print("End of training: %s" % end_of_training)
        datetime_indices = pd.date_range(
            specs.start_of_training, end_of_training, freq="15T"
        )
    elif time_range == "test":
        start_of_testing = get_closest_quarter(
            specs.start_of_training
            + (length_of_data * specs.ratio_training_testing_data)
            + timedelta(minutes=15)
        )
        # print("Start of testing: %s" % start_of_testing)
        # print("End of testing: %s" % specs.end_of_testing)
        datetime_indices = pd.date_range(
            start_of_testing, specs.end_of_testing, freq="15T"
        )

    return datetime_indices


def lag_to_suffix(lag: int) -> str:
    """
    Return the suffix for a column, given its lag.
    """
    if lag < 0:
        str_lag = "f" + str(abs(lag))
    else:
        str_lag = "l" + str(abs(lag))
    return str_lag


def add_lags(
    df: pd.DataFrame, column: str, lags: List[int], name_as: str = "quarter_hour"
) -> Optional[pd.DataFrame]:
    """
    Creates lag columns for a column in the dataframe. Lags are in fifteen minute steps (15T).
    Positive values are lags, while negative values are future values.
    The new columns are named like the lagged column, plus "_l<lag>" (or "_f<lag>" for positive lags (future values)),
    where <lag> is the 15-minute lag value or the translation of it into days, weeks or years.
    In case of positive 'lags' (future values), new columns are named like the lagged column, plus "_f<lag>".

    TODO: We could also review if using statsmodels.tsa.tsatools.add_lag is of interest, but here self-made is probably
          what we need.
    """

    if not lags:
        return df

    # Make sure the DataFrame has rows to accommodate each lag
    max_lag = timedelta(minutes=15 * max(lags))
    min_lag = timedelta(minutes=15 * min(lags))
    df_start = min(df.index[0], df.index[0] + min_lag)
    df_end = max(df.index[-1], df.index[-1] + max_lag)
    df = df.reindex(
        pd.date_range(
            start=df_start.astimezone(pytz.utc),
            end=df_end.astimezone(pytz.utc),
            freq=df.index.freq,
        )
    )

    lag_names = [l for l in lags]
    if name_as == "hour":
        lag_names = [int(l / 4) for l in lags]
    if name_as == "day":
        lag_names = [int(l / 96) for l in lags]
    if name_as == "week":
        lag_names = [int(l / 96 / 7) for l in lags]
    if name_as == "year":
        lag_names = [int(l / 96 / 365) for l in lags]

    for lag in lags:
        lag_name = lag_names[lags.index(lag)]
        df[column + "_" + lag_to_suffix(lag_name)] = df[column].shift(lag)

    return df



"""Exceptions that are worth catching"""


class MissingData(Exception):
    """Data is missing where they needs to be some."""

    pass


class UnsupportedModel(Exception):
    """If a statistic model is not implemented in this tool (yet)."""

    pass


class ModelLocationProblem(Exception):
    """If a serialised model cannot be found."""

    pass



from typing import Tuple
from datetime import datetime, timedelta

from pandas import DataFrame, Series, date_range
import pytz

from ts_forecasting_pipeline import MODEL_CLASSES, ModelState, ModelSpecs
from ts_forecasting_pipeline.modelling import create_fitted_model
from ts_forecasting_pipeline.featuring import construct_features, get_time_steps
from ts_forecasting_pipeline.utils.time_utils import get_closest_quarter


"""
Functionality for making predictions per time slot.
"""


def make_forecast_for(
    specs: ModelSpecs, features: DataFrame, model: MODEL_CLASSES
) -> float:
    """
    Make a forecast for the given feature vector.
    """

    y_hat = model.predict(features)

    # Apply back-transformation if the output data was transformed
    if specs.transformation is not None:
        y_hat = specs.transformation.back_transform(y_hat)

    return y_hat


def update_model(
    time_step: datetime,
    current_model: MODEL_CLASSES,
    specs: ModelSpecs,
    feature_frame: DataFrame,
) -> ModelState:
    new_model: MODEL_CLASSES = current_model
    """ Create model if current one is outdated or not yet created."""
    if (
        current_model is None
        or time_step - specs.creation_time >= specs.remodel_frequency
    ):
        # print("Fitting new model before predicting %s ..." % time_step)
        if current_model is not None:
            # move the model's series specs further in time
            specs.start_of_training = specs.start_of_training + specs.remodel_frequency
            specs.end_of_testing = specs.end_of_testing + specs.remodel_frequency
        relevant_time_steps = get_time_steps(time_range="train", specs=specs)
        new_model = create_fitted_model(
            specs, "", regression_frame=feature_frame.loc[relevant_time_steps]
        )
        specs.creation_time = time_step

    return ModelState(new_model, specs)


def make_rolling_forecasts(
    start: datetime,  # Start of forecast period
    end: datetime,  # End of forecast period
    model_specs: ModelSpecs,
) -> Tuple[Series, ModelState]:
    """
    Repeatedly call make_forecast - for all time steps the desired time window.
    The time window of the specs (training + test data) is moved forwards also step by step.
    Will fail if series specs do not allocate enough data.
    May create a model whenever the previous one is outdated.
    Return Pandas.Series as result, as well as the last ModelState.
    """

    # Prepare time range
    for dt in (start, end):
        if dt.tzinfo is None:
            dt.replace(tzinfo=pytz.utc)
    training_start = get_closest_quarter(model_specs.start_of_training)
    start = get_closest_quarter(start)
    end = get_closest_quarter(end)

    # First, compute one big feature frame, once.
    feature_frame: DataFrame = construct_features((training_start, end), model_specs)

    """
    specs.outcome_var = ObjectSeriesSpecs(
        name=specs.outcome_var.name, data=specs.outcome_var.load_series()
    )
    specs.regressors = [
        ObjectSeriesSpecs(name=r.name, data=r.load_series()) for r in specs.regressors
    ]
    """
    values = Series(
        index=date_range(start, end, freq="15T", closed="left", tz=start.tzinfo)
    )
    time_step = start
    model = None
    print("[ts-forecasting-pipeline] Forecasting from %s to %s" % (start, end))
    while time_step < end:
        model, specs = update_model(
            time_step, model, model_specs, feature_frame=feature_frame
        ).split()
        features = feature_frame.loc[time_step:time_step].iloc[:, 1:]
        values[time_step] = make_forecast_for(model_specs, features, model)
        time_step = time_step + timedelta(minutes=15)

    return values, ModelState(model, model_specs)



from typing import Dict, List, Optional, Tuple, Type, Union
from datetime import datetime, timedelta, tzinfo
from pprint import pformat
import json
import warnings

import pytz
import dateutil.parser
import numpy as np
import pandas as pd
from statsmodels.base.transform import BoxCox
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query

from sqlalchemy.dialects import postgresql
from ts_forecasting_pipeline.utils.debug_utils import render_query
from ts_forecasting_pipeline.utils.time_utils import tz_aware_utc_now


DEFAULT_RATIO_TRAINING_TESTING_DATA = 2 / 3
DEFAULT_REMODELING_FREQUENCY = timedelta(days=1)

np.seterr(all="warn")
warnings.filterwarnings("error", message="invalid value encountered in power")


class SeriesSpecs(object):
    """Describes a time series (e.g. a pandas Series).
    In essence, a column in the regression frame, filled with numbers.

    Using this class, the column will be filled with NaN values.

    If you have data to be loaded in automatically, you should be using one of the subclasses, which allow to describe
    or pass in an actual data source to be loaded.

    When dealing with columns, our code should usually refer to this superclass so it does not need to care
    which kind of data source it is dealing with.
    """

    # The name in the resulting feature frame, and possibly in the saved model specs (named by outcome var)
    name: str
    # The name in the data source, if source is a pandas DataFrame or database Table - if None, the name will be tried
    column: Optional[str]
    # timezone of the data - useful when de-serializing (e.g. pandas serialises to UTC)
    original_tz: tzinfo

    def __init__(self, name: str, column: str = None, original_tz: tzinfo = None):
        self.name = name
        self.column = column
        if self.column is None:
            self.column = name
        self.original_tz = original_tz
        self.__series_type__ = self.__class__.__name__

    def as_dict(self):
        return vars(self)

    def load_series(self) -> pd.Series:
        return pd.Series()

    def __repr__(self):
        return "%s: <%s>" % (self.__class__.__name__, self.as_dict())


class ObjectSeriesSpecs(SeriesSpecs):
    """
    Spec for a pd.Series object that is being passed in and is stored directly in the specs.
    The data is not mutatable after creation.
    Note: The column argument is not used, as the series has only one column.
    """

    data: pd.Series

    def __init__(
        self, data: pd.Series, name: str, column: str = None, original_tz: tzinfo = None
    ):
        super().__init__(name, None, original_tz)
        self.data = data
        self.original_tz = data.index.tzinfo
        if self.original_tz is None:
            self.original_tz = pytz.utc

    def load_series(self) -> pd.Series:
        return self.data


class DFFileSeriesSpecs(SeriesSpecs):
    """
    Spec for a pandas DataFrame source.
    This class holds the filename, from which we unpickle the data frame, then read the column.
    """

    file_path: str

    def __init__(
        self, file_path: str, name: str, column: str = None, original_tz: tzinfo = None
    ):
        super().__init__(name, column, original_tz)
        self.file_path = file_path

    def load_series(self) -> pd.Series:
        df: pd.DataFrame = pd.read_pickle(self.file_path)
        if df.index.tzinfo is None:
            self.original_tz = pytz.utc
            df.index = df.index.tz_localize(self.original_tz)
        else:
            self.original_tz = df.index.tzinfo
        return df[self.column]


class DBSeriesSpecs(SeriesSpecs):

    """Define how to query a database for time series values.
    This works via a SQLAlchemy query.
    This query should return the needed information for the forecasting pipeline:
    A "datetime" column (which will be set as index) and the values column (named by name or column,
    see SeriesSpecs.__init__, defaults to "value"). For example:
    TODO: show an example"""

    db: Engine
    query: Query

    def __init__(
        self,
        db_engine: Engine,
        query: Query,
        name: str = "value",
        column: str = None,
        original_tz: tzinfo = pytz.utc,  # postgres stores naive datetimes
    ):
        super().__init__(name, column, original_tz)
        self.db_engine = db_engine
        self.query = query

    def load_series(self) -> pd.Series:
        print(
            "[ts-forecasting-pipeline] Reading %s data from database"
            % self.query.column_descriptions[0]["entity"].__tablename__
        )
        """
        from sqlalchemy.dialects import postgresql
        cq = self.query.statement.compile(dialect=postgresql.dialect())
        print("Query: %s" % str(cq))
        print("Params: %s" % str(cq.params))
        """

        series_orig = pd.read_sql(self.query.statement, self.db_engine)
        if series_orig.empty:
            # e.g. during a test session, when not yet committed, so we cannot go via engine.
            # TODO: make enforceable by a param or always go this route (get columns from query)
            series_orig = pd.DataFrame.from_records(
                self.query.all(), columns=list(series_orig.columns)
            )
        series_orig["datetime"] = pd.to_datetime(series_orig["datetime"], utc=True)

        # Raise error if data is empty or contains nan values
        if series_orig.empty:
            raise ValueError(
                "No values found in database for the requested %s data. It's no use to continue I'm afraid."
                " Here's a print-out of the database query:\n\n%s\n\n"
                % (
                    self.query.column_descriptions[0]["entity"].__tablename__,
                    render_query(self.query.statement, dialect=postgresql.dialect()),
                )
            )
        if series_orig.isnull().values.any():
            raise ValueError(
                "Nan values found in database for the requested %s data. It's no use to continue I'm afraid."
                " Here's a print-out of the database query:\n\n%s\n\n"
                % (
                    self.query.column_descriptions[0]["entity"].__tablename__,
                    render_query(self.query.statement, dialect=postgresql.dialect()),
                )
            )

        # Keep the most recent observation
        series_orig = (
            series_orig.sort_values(by=["horizon"], ascending=True)
            .drop_duplicates(subset=["datetime"], keep="first")
            .sort_values(by=["datetime"])
        )
        series_orig.set_index("datetime", drop=True, inplace=True)

        if series_orig.index.tzinfo is None:
            if self.original_tz is not None:
                series_orig.index = series_orig.index.tz_localize(self.original_tz)
        else:
            series_orig.index = series_orig.index.tz_convert(self.original_tz)
        series = series_orig.resample("15T").mean()
        return series["value"]


class Transformation(object):
    """Base class for transformations.
    Initialise with your custom transformation parameters and define custom functions to transform and back-transform.
    """

    def __init__(self, **kwargs):
        """Initialise transformation with named parameters.
        For example:
            >>> transformation = Transformation(lambda1=0.5, lambda2=1)
            >>> transformation.params.lambda1
            0.5
        """
        self.params = type("Params", (), {})
        self._set_params(**kwargs)

    def _set_params(self, **kwargs):
        """Assign named variables as attributes."""
        for k, v in kwargs.items():
            setattr(self.params, k, v)

    def transform(self, x: np.array) -> Tuple[np.array, dict]:
        """Return transformed data and set new transformation parameters if applicable."""
        params = {}
        y = x
        self._set_params(**params)
        return y, params

    def back_transform(self, x: np.array) -> np.array:
        """Return back-transformed data."""
        y = x
        return y


class BoxCoxTransformation(Transformation):
    """Box-Cox transformation.

     For positive-only or negative-only data, no parameters are needed.
     For non-negative or non-positive data with zero values, set lambda2 to a positive number (e.g. 1).

                            {   ( (x' + lambda2) ^ lambda1 − 1) / lambda1        if lambda1 != 0
     y(lambda1, lambda2) = {
                            {   log(x' + lambda2)                                if lambda1 == 0

    where:            x' = x * lambda3
    """

    def __init__(self, lambda2: float = 0.1):
        super().__init__(lambda2=lambda2)

    def transform(
        self, x: Union[np.array, pd.DataFrame, pd.Series]
    ) -> Tuple[np.array, dict]:
        params = {}
        if isinstance(x, (pd.DataFrame, pd.Series)):
            x = np.array(x.values, dtype=np.float64)

        if (x[~np.isnan(x)] + self.params.lambda2 > 0).all():
            y, params["lambda1"] = BoxCox.transform_boxcox(
                BoxCox(), x + self.params.lambda2
            )
            params["lambda3"] = 1
        elif (x[~np.isnan(x)] - self.params.lambda2 < 0).all():
            y, params["lambda1"] = BoxCox.transform_boxcox(
                BoxCox(), -x + self.params.lambda2
            )
            params["lambda3"] = -1
        else:
            raise ValueError(
                "Box-Cox transformation not suitable for x with both positive and negative values."
            )
        self._set_params(**params)
        return y

    def back_transform(self, x: np.array) -> np.array:
        try:
            y = (
                BoxCox.untransform_boxcox(BoxCox(), x, lmbda=self.params.lambda1)
                - self.params.lambda2
            ) / self.params.lambda3
        except Warning as w:
            if (
                w.__str__() == "invalid value encountered in power"
                and (x < 0).all()
                and self.params.lambda1 < 1
            ):

                # Resolve a numpy problem for raising a number close to 0 to a large number, i.e. -0.12^6.25
                y = (np.zeros(*x.shape) - self.params.lambda2) / self.params.lambda3
            else:
                print(
                    "Back-transform failed for y(x, lambda1, lambda2, lambda3) with:\n"
                    "x = %s\n"
                    "lambda1 = %s\n"
                    "lambda2 = %s\n"
                    "lambda3 = %s\n"
                    "warning = %s\n"
                    "Returning 0 value instead."
                    % (
                        x,
                        self.params.lambda1,
                        self.params.lambda2,
                        self.params.lambda3,
                        w,
                    )
                )
                y = (np.zeros(*x.shape) - self.params.lambda2) / self.params.lambda3
        return y


class ModelSpecs(object):
    """Describes a model and how it was trained.
    """

    outcome_var: SeriesSpecs
    model_type: Type  # e.g. statsmodels.api.OLS, sklearn.linear_model.LinearRegression, ...
    model_params: dict
    horizon: timedelta
    lags: List[int]
    regressors: List[SeriesSpecs]
    # Start of training data set
    start_of_training: datetime
    # End of testing data set
    end_of_testing: datetime
    # This determines the cutoff point between training and testing data
    ratio_training_test_data: float
    # time this model was created, defaults to UTC now
    creation_time: datetime
    model_filename: str
    remodel_frequency: timedelta
    # Custom transformation to perform on the outcome data. Called after relevant SeriesSpecs were resolved.
    transformation: Transformation
    # Custom transformation to perform on each regressor data. Called after relevant SeriesSpecs were resolved.
    regressor_transformation: Dict[str, Transformation]

    def __init__(
        self,
        outcome_var: Union[str, SeriesSpecs, pd.Series],
        model: Union[
            Type, Tuple[Type, dict]
        ],  # Model class and optionally initialization parameters
        start_of_training: Union[str, datetime],
        end_of_testing: Union[str, datetime],
        horizon: timedelta,
        lags: List[int] = None,
        regressors: Union[List[str], List[SeriesSpecs], List[pd.Series]] = None,
        ratio_training_testing_data=DEFAULT_RATIO_TRAINING_TESTING_DATA,
        remodel_frequency: Union[str, timedelta] = DEFAULT_REMODELING_FREQUENCY,
        model_filename: str = None,
        creation_time: Union[str, datetime] = None,
        transformation: Transformation = None,
        regressor_transformation: Dict[str, Transformation] = None,
    ):
        """Create a ModelSpecs instance. Accepts all parameters as string (besides transform - TODO) for
         deserialization support (JSON strings for all parameters which are not natively JSON-parseable,)"""
        self.outcome_var = parse_series_specs(outcome_var, "y")
        self.model_type = model[0] if isinstance(model, tuple) else model
        self.model_params = model[1] if isinstance(model, tuple) else {}
        self.horizon = horizon
        self.lags = lags
        if self.lags is None:
            self.lags = []
        if regressors is None:
            self.regressors = []
        else:
            self.regressors = [
                parse_series_specs(r, "Regressor%d" % (regressors.index(r) + 1))
                for r in regressors
            ]
        if isinstance(start_of_training, str):
            self.start_of_training = dateutil.parser.parse(start_of_training)
        else:
            self.start_of_training = start_of_training
        if isinstance(end_of_testing, str):
            self.end_of_testing = dateutil.parser.parse(end_of_testing)
        else:
            self.end_of_testing = end_of_testing
        self.ratio_training_testing_data = ratio_training_testing_data
        if isinstance(creation_time, str):
            self.creation_time = dateutil.parser.parse(creation_time)
        elif creation_time is None:
            self.creation_time = tz_aware_utc_now()
        else:
            self.creation_time = creation_time
        self.model_filename = model_filename
        if isinstance(remodel_frequency, str):
            self.remodel_frequency = timedelta(
                days=int(remodel_frequency) / 60 / 60 / 24
            )
        else:
            self.remodel_frequency = remodel_frequency
        self.transformation = transformation
        self.regressor_transformation = regressor_transformation

    def as_dict(self):
        return vars(self)

    def __repr__(self):
        return "ModelSpecs: <%s>" % pformat(vars(self))


def parse_series_specs(
    specs: Union[str, SeriesSpecs, pd.Series], name: str = None
) -> SeriesSpecs:
    if isinstance(specs, str):
        return load_series_specs_from_json(specs)
    elif isinstance(specs, pd.Series):
        return ObjectSeriesSpecs(specs, name)
    else:
        return specs


def load_series_specs_from_json(s: str) -> SeriesSpecs:
    json_repr = json.loads(s)
    series_class = globals()[json_repr["__series_type__"]]
    if series_class == ObjectSeriesSpecs:
        # load pd.Series from string, will be UTC-indexed, so apply original_tz
        json_repr["data"] = pd.read_json(
            json_repr["data"], typ="series", convert_dates=True
        )
        json_repr["data"].index = json_repr["data"].index.tz_localize(
            json_repr["original_tz"]
        )
    return series_class(
        **{k: v for k, v in json_repr.items() if not k.startswith("__")}
    )
